
#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "OrbiterSoundSDK3.h"
const double SWATH_FUELMASS = 5000.0;
const double SWATH_ISP = 6.25e6;
const double SWATH_MAXMAINTH = 1.2e5;
const double SWATH_MAXRETROTH = 1e5;
const double SWATH_MAXRCSTH = 1e5;


// ==============================================================
// Airfoil coefficient functions
// Return lift, moment and zero-lift drag coefficients as a
// function of angle of attack (alpha or beta)
// ==============================================================

// 1. vertical lift component (wings and body)

void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD,-2*RAD, 15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nabsc]  = {       0,      0,   -0.4,     0,    0.7,     1,   0.8,     0,      0};
	static const double CM[nabsc]  = {       0,      0,  0.014, 0.004,-0.0074,-0.094,-0.012,     0,      0};
	for (int i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = CM[i] + (CM[i+1]-CM[i]) * f;  // aoa-dependent moment coefficient
	double saoa = sin(aoa);
	double pd = 0.015 + 0.4*saoa*saoa;  // profile drag
	*cd = pd + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	// profile drag + (lift-)induced drag + transonic/supersonic wave (compressibility) drag
}

// 2. horizontal lift component (vertical stabilisers and body)

void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {       0,    +0.3,      0,   -0.3,  +0.3,     0,   -0.3,      0};
	for (int i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.015 + oapiGetInducedDrag (*cl, 1.5, 0.6) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}

class SWATH: public VESSEL2 {
public:
	SWATH (OBJHANDLE hVessel, int flightmodel)
		: VESSEL2 (hVessel, flightmodel) {}
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep (double simt, double simdt, double mjd);
	void clbkPostCreation(void);
	int Sailor;
	
private:
MESHHANDLE exmesh_tpl;
double Play_Swath1;


    VECTOR3 RelPos1;        
    VECTOR3 MinDist1;
     OBJHANDLE hbase;
   
	
};


void SWATH::clbkSetClassCaps (FILEHANDLE cfg)
{
	THRUSTER_HANDLE th_main, th_retro, th_hover, th_rcs[14], th_group[4];
  //  SURFHANDLE tex1 = oapiRegisterExhaustTexture ("SwathExhaust");
	SetSize (50);
	SetEmptyMass (10000);
	SetCW (1.0, 1.0, 1.8, 1.8);
	SetReentryTexture (NULL);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetCrossSections (_V(1000,1000,1000));
	SetRotDrag (_V(0.6,0.6,0.35));
	if (GetFlightModel() >= 1) {
		SetPitchMomentScale (5e-5);
		SetBankMomentScale (5e-5);
	}
	SetPMI (_V(80,90,17.5));
	SetTrimScale (0.05);
	SetCameraOffset (_V(0,12.7,20));
	SetSurfaceFrictionCoeff (0.26, 0.6);
	SetTouchdownPoints (_V(0,-9.5,20), _V(-10,-9.5,-20), _V(10,-9.5,-20));
	Play_Swath1 = 0 ;

	// propellant resources
	PROPELLANT_HANDLE hpr;
	hpr = CreatePropellantResource (SWATH_FUELMASS);
	
PARTICLESTREAMSPEC exhaust_main = {
		0, 5.0, 5, 12, 3, 6, 2, 10.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	};
// ********************* aerodynamics ***********************

	CreateAirfoil (LIFT_VERTICAL, _V(0,0,-0.2), VLiftCoeff, 5, 1, 1.5);
	// wing and body lift+drag components

	CreateAirfoil (LIFT_HORIZONTAL, _V(0,0,-4), HLiftCoeff, 5, 500, 1.5);
	// vertical stabiliser and body lift and drag components

	
	CreateControlSurface (AIRCTRL_RUDDER,       80, 1.5, _V(   0,0,-25), AIRCTRL_AXIS_YPOS);
	SetMeshVisibilityMode (AddMesh (exmesh_tpl = oapiLoadMeshGlobal ("Swath")), MESHVIS_ALWAYS);
 
	// ***************** thruster definitions *******************
     
     
	th_main = CreateThruster (_V(0,0,-4.5), _V(0,0,1), SWATH_MAXMAINTH, hpr, SWATH_ISP);
	CreateThrusterGroup (&th_main, 1, THGROUP_MAIN);
	//AddExhaust (th_main, 44, 3, _V(-18.5,-16,-57.5), _V(0,0,-1),tex1);
    //AddExhaust (th_main, 44, 3, _V(18.5,-16,-57.5), _V(0,0,-1),tex1);
    AddExhaustStream (th_main, _V(-18.5,-16,-57.5), &exhaust_main);
	AddExhaustStream (th_main, _V(18.5,-16,-57.5), &exhaust_main);
	th_retro = CreateThruster (_V(0,0,-4.5), _V(0,0,-1), SWATH_MAXRETROTH, hpr, SWATH_ISP);
	CreateThrusterGroup (&th_retro, 1, THGROUP_RETRO);
	//AddExhaust (th_retro, 44, 3, _V(-18.5,-16,57.5), _V(0,0,1),tex1);
   // AddExhaust (th_retro, 44, 3, _V(18.5,-16,57.5), _V(0,0,1),tex1);
	
	
    

	th_rcs[ 0] = CreateThruster (_V( 1,0, 3), _V(0,1,0), SWATH_MAXRCSTH/1000, hpr, SWATH_ISP);
	th_rcs[ 1] = CreateThruster (_V( 1,0, 3), _V(0,-1,0), SWATH_MAXRCSTH/1000, hpr, SWATH_ISP);
	th_rcs[ 2] = CreateThruster (_V(-1,0, 3), _V(0,1,0), SWATH_MAXRCSTH/1000, hpr, SWATH_ISP);
	th_rcs[ 3] = CreateThruster (_V(-1,0, 3), _V(0,-1,0), SWATH_MAXRCSTH/1000, hpr, SWATH_ISP);
	th_rcs[ 4] = CreateThruster (_V( 1,0,-3), _V(0,1,0), SWATH_MAXRCSTH/1000, hpr, SWATH_ISP);
	th_rcs[ 5] = CreateThruster (_V( 1,0,-3), _V(0,-1,0), SWATH_MAXRCSTH/1000, hpr, SWATH_ISP);
	th_rcs[ 6] = CreateThruster (_V(-1,0,-3), _V(0,1,0), SWATH_MAXRCSTH/1000, hpr, SWATH_ISP);
	th_rcs[ 7] = CreateThruster (_V(-1,0,-3), _V(0,-1,0), SWATH_MAXRCSTH/1000, hpr, SWATH_ISP);
	th_rcs[ 8] = CreateThruster (_V( 1,0, 3), _V(-1,0,0), SWATH_MAXRCSTH, hpr, SWATH_ISP);
	th_rcs[ 9] = CreateThruster (_V(-1,0, 3), _V(1,0,0), SWATH_MAXRCSTH, hpr, SWATH_ISP);
	th_rcs[10] = CreateThruster (_V( 1,0,-3), _V(-1,0,0), SWATH_MAXRCSTH, hpr, SWATH_ISP);
	th_rcs[11] = CreateThruster (_V(-1,0,-3), _V(1,0,0), SWATH_MAXRCSTH, hpr, SWATH_ISP);
	th_rcs[12] = CreateThruster (_V( 0,0,-3), _V(0,0,1), SWATH_MAXRCSTH, hpr, SWATH_ISP);
	th_rcs[13] = CreateThruster (_V( 0,0, 3), _V(0,0,-1), SWATH_MAXRCSTH, hpr, SWATH_ISP);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[2];
	th_group[2] = th_rcs[5];
	th_group[3] = th_rcs[7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[3];
	th_group[2] = th_rcs[4];
	th_group[3] = th_rcs[6];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[4];
	th_group[2] = th_rcs[3];
	th_group[3] = th_rcs[7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[5];
	th_group[2] = th_rcs[2];
	th_group[3] = th_rcs[6];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[4];
	th_group[2] = th_rcs[2];
	th_group[3] = th_rcs[6];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_UP);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[5];
	th_group[2] = th_rcs[3];
	th_group[3] = th_rcs[7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_DOWN);

	th_group[0] = th_rcs[8];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[9];
	th_group[1] = th_rcs[10];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[8];
	th_group[1] = th_rcs[10];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_LEFT);

	th_group[0] = th_rcs[9];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_RIGHT);

	CreateThrusterGroup (th_rcs+12, 1, THGROUP_ATT_FORWARD);
	CreateThrusterGroup (th_rcs+13, 1, THGROUP_ATT_BACK);

	
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new SWATH (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (SWATH*)vessel;
}

void SWATH::clbkPostCreation (void)
{
	Sailor=ConnectToOrbiterSoundDLL3(GetHandle());

	ReplaceStockSound3(Sailor,"Sound\\Rho1\\Swath\\Swathengines.wav",	REPLACE_MAIN_THRUST);
	ReplaceStockSound3(Sailor,"Sound\\Rho1\\Swath\\SwathAMB1.wav",	REPLACE_COCKPIT_AMBIENCE_1);
    ReplaceStockSound3(Sailor,"Sound\\Rho1\\Swath\\SwathAMB2.wav",	REPLACE_COCKPIT_AMBIENCE_2);
	ReplaceStockSound3(Sailor,"Sound\\Rho1\\Swath\\SwathAMB3.wav",	REPLACE_COCKPIT_AMBIENCE_3);
	ReplaceStockSound3(Sailor,"Sound\\Rho1\\Swath\\SwathAMB4.wav",	REPLACE_COCKPIT_AMBIENCE_4);
	RequestLoadVesselWave3(Sailor,1,"Sound\\Rho1\\Swath\\Greetings5.wav",DEFAULT);
	SoundOptionOnOff3(Sailor,PLAYCOUNTDOWNWHENTAKEOFF,FALSE);
	SoundOptionOnOff3(Sailor,DISPLAYTIMER,FALSE);
	SoundOptionOnOff3(Sailor,PLAYRADIOATC,FALSE);
	SoundOptionOnOff3(Sailor,PLAYATTITUDETHRUST,FALSE);
	SoundOptionOnOff3(Sailor,PLAYGPWS,FALSE);
	
}


void SWATH::clbkPreStep (double simt, double simdt, double mjd) 
{
 ///////////////////////////////////////////////////////////////////
       // position relative to object
      VESSEL *v = oapiGetFocusInterface();// to determine which vehicle has focus  
      MinDist1 =  _V(750,750,750); //Distance to trigger sound
      hbase=oapiGetVesselByName("pip");// name of vessel to detect distance (shows in F3 list)
	     if (oapiIsVessel(hbase) == true) { // check if vessel exist
		  v->GetRelativePos (hbase, RelPos1);//crashes if hvessel isnt in the senario
	  		                                     //
// change numbers to positive
       if (RelPos1.x < 0.0){
 	RelPos1.x = -(RelPos1.x) ;}
       if (RelPos1.y < 0.0){
    RelPos1.y = -(RelPos1.y) ;}
       if (RelPos1.z < 0.0){ 
    RelPos1.z = -(RelPos1.z) ;} 
   






     if (MinDist1.x > RelPos1.x && MinDist1.y > RelPos1.y && MinDist1.z > RelPos1.z){

		 if (Play_Swath1 < 1){ // play once !
    
			 PlayVesselRadioExclusiveWave3(Sailor,1,255); 
			 Play_Swath1 = 1;


		 }// play once
	 }

	
		 } }