
#define STRICT
#define ORBITER_MODULE

#include "orbitersdk.h"
#include "OrbiterSoundSDK3.h"
const double LNCH_FUELMASS = 500.0;
const double LNCH_ISP = 6.25e4;
const double LNCH_MAXMAINTH = 5.5e3;
const double LNCH_MAXRCSTH = 1e3;



// ==============================================================
// Airfoil coefficient functions
// Return lift, moment and zero-lift drag coefficients as a
// function of angle of attack (alpha or beta)
// ==============================================================

// 1. vertical lift component (wings and body)

void VLiftCoeff (double aoa, double M, double Re, double *cl, double *cm, double *cd)
{
	const int nabsc = 9;
	static const double AOA[nabsc] = {-180*RAD,-60*RAD,-30*RAD,-2*RAD, 15*RAD,20*RAD,25*RAD,60*RAD,180*RAD};
	static const double CL[nabsc]  = {       0,      0,   -0.4,     0,    0.7,     1,   0.8,     0,      0};
	static const double CM[nabsc]  = {       0,      0,  0.014, 0.004,-0.0074,-0.094,-0.012,     0,      0};
	for (int i = 0; i < nabsc-1 && AOA[i+1] < aoa; i++);
	double f = (aoa-AOA[i]) / (AOA[i+1]-AOA[i]);
	*cl = CL[i] + (CL[i+1]-CL[i]) * f;  // aoa-dependent lift coefficient
	*cm = CM[i] + (CM[i+1]-CM[i]) * f;  // aoa-dependent moment coefficient
	double saoa = sin(aoa);
	double pd = 0.015 + 0.4*saoa*saoa;  // profile drag
	*cd = pd + oapiGetInducedDrag (*cl, 1.5, 0.7) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
	// profile drag + (lift-)induced drag + transonic/supersonic wave (compressibility) drag
}

// 2. horizontal lift component (vertical stabilisers and body)

void HLiftCoeff (double beta, double M, double Re, double *cl, double *cm, double *cd)
{
	const int nabsc = 8;
	static const double BETA[nabsc] = {-180*RAD,-135*RAD,-90*RAD,-45*RAD,45*RAD,90*RAD,135*RAD,180*RAD};
	static const double CL[nabsc]   = {       0,    +0.3,      0,   -0.3,  +0.3,     0,   -0.3,      0};
	for (int i = 0; i < nabsc-1 && BETA[i+1] < beta; i++);
	*cl = CL[i] + (CL[i+1]-CL[i]) * (beta-BETA[i]) / (BETA[i+1]-BETA[i]);
	*cm = 0.0;
	*cd = 0.015 + oapiGetInducedDrag (*cl, 1.5, 0.6) + oapiGetWaveDrag (M, 0.75, 1.0, 1.1, 0.04);
}

class LNCH: public VESSEL2 {
public:
	LNCH (OBJHANDLE hVessel, int flightmodel)
		: VESSEL2 (hVessel, flightmodel) {}
	void clbkSetClassCaps (FILEHANDLE cfg);
	void clbkPreStep (double simt, double simdt, double mjd);
	void clbkPostCreation(void);
	int Fish;
		
private:
double Play_Launch1;
MESHHANDLE exmesh_tpl;

    VECTOR3 RelPos1;
    VECTOR3 RelPos2;
    VECTOR3 RelPos3;         
    VECTOR3 MinDist1;
     OBJHANDLE hvessel;
};


void LNCH::clbkSetClassCaps (FILEHANDLE cfg)
{
	THRUSTER_HANDLE th_main, th_hover, th_rcs[14], th_group[4];

	SetSize (5);
	SetEmptyMass (1000);
	SetCW (0.09, 1.0, 1.8, 1.8);
	SetReentryTexture (NULL);
	SetWingAspect (0.7);
	SetWingEffectiveness (2.5);
	SetCrossSections (_V(19,37,8));
	SetRotDrag (_V(0.6,0.6,0.35));
	if (GetFlightModel() >= 1) {
		SetPitchMomentScale (5e-5);
		SetBankMomentScale (5e-5);
	}
	SetPMI (_V(8,9,1.75));
	SetTrimScale (0.05);
	SetCameraOffset (_V(0,0.5,4.5));
	
	SetTouchdownPoints (_V(0,-1.35,1.2), _V(-1.53,-1.11,-4.83), _V(1.53,-1.11,-4.83));
	Play_Launch1 = 0 ;
    SetMeshVisibilityMode (AddMesh (exmesh_tpl = oapiLoadMeshGlobal ("Orca")), MESHVIS_ALWAYS);
	// propellant resources
	PROPELLANT_HANDLE hpr;
	hpr = CreatePropellantResource (LNCH_FUELMASS);
	

// ********************* aerodynamics ***********************

	CreateAirfoil (LIFT_VERTICAL, _V(0,0,-0.2), VLiftCoeff, 5, 40, 1.5);
	// wing and body lift+drag components

	CreateAirfoil (LIFT_HORIZONTAL, _V(0,0,-4), HLiftCoeff, 5, 15, 1.5);
	// vertical stabiliser and body lift and drag components

	CreateControlSurface (AIRCTRL_ELEVATOR,     1.2, 1.5, _V(   0,0,-7.2), AIRCTRL_AXIS_XPOS);
	CreateControlSurface (AIRCTRL_RUDDER,       0.8, 1.5, _V(   0,0,-7.2), AIRCTRL_AXIS_YPOS);
	CreateControlSurface (AIRCTRL_AILERON,      0.2, 1.5, _V( 7.5,0,-7.2), AIRCTRL_AXIS_XPOS);
	CreateControlSurface (AIRCTRL_AILERON,      0.2, 1.5, _V(-7.5,0,-7.2), AIRCTRL_AXIS_XNEG);
	CreateControlSurface (AIRCTRL_ELEVATORTRIM, 0.1, 1.5, _V(   0,0,-7.2), AIRCTRL_AXIS_XPOS);

	// ***************** thruster definitions *******************
     
      PARTICLESTREAMSPEC contrail_main = {
		0, 5.0, 16, 200, 0.15, 1.0, 5, 3.0, PARTICLESTREAMSPEC::DIFFUSE,
		PARTICLESTREAMSPEC::LVL_PSQRT, 0, 2,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-4, 1
	};
	PARTICLESTREAMSPEC exhaust_main = {
		0, 2.0, 20, 200, 0.05, 0.1, 8, 1.0, PARTICLESTREAMSPEC::EMISSIVE,
		PARTICLESTREAMSPEC::LVL_SQRT, 0, 1,
		PARTICLESTREAMSPEC::ATM_PLOG, 1e-5, 0.1
	};
	
	
	th_main = CreateThruster (_V(0,0,-4.5), _V(0,0,1), LNCH_MAXMAINTH, hpr, LNCH_ISP);
	CreateThrusterGroup (&th_main, 1, THGROUP_MAIN);
	AddExhaust (th_main, 4, .2, _V(-.96,0,-5.13), _V(0,0,-1));
      AddExhaust (th_main, 4, .2, _V(.96,0,-5.13), _V(0,0,-1));

	
	AddExhaustStream (th_main, _V(-.96,0,-5.13), &contrail_main);
    AddExhaustStream (th_main, _V(-.96,0,-5.13), &contrail_main);
	AddExhaustStream (th_main, _V(.96,0,-5.13), &exhaust_main);
	AddExhaustStream (th_main, _V(.96,0,-5.13), &exhaust_main);

	th_rcs[ 0] = CreateThruster (_V( 1,0, 3), _V(0,1,0), LNCH_MAXRCSTH, hpr, LNCH_ISP);
	th_rcs[ 1] = CreateThruster (_V( 1,0, 3), _V(0,-1,0), LNCH_MAXRCSTH, hpr, LNCH_ISP);
	th_rcs[ 2] = CreateThruster (_V(-1,0, 3), _V(0,1,0), LNCH_MAXRCSTH, hpr, LNCH_ISP);
	th_rcs[ 3] = CreateThruster (_V(-1,0, 3), _V(0,-1,0), LNCH_MAXRCSTH, hpr, LNCH_ISP);
	th_rcs[ 4] = CreateThruster (_V( 1,0,-3), _V(0,1,0), LNCH_MAXRCSTH, hpr, LNCH_ISP);
	th_rcs[ 5] = CreateThruster (_V( 1,0,-3), _V(0,-1,0), LNCH_MAXRCSTH, hpr, LNCH_ISP);
	th_rcs[ 6] = CreateThruster (_V(-1,0,-3), _V(0,1,0), LNCH_MAXRCSTH, hpr, LNCH_ISP);
	th_rcs[ 7] = CreateThruster (_V(-1,0,-3), _V(0,-1,0), LNCH_MAXRCSTH, hpr, LNCH_ISP);
	th_rcs[ 8] = CreateThruster (_V( 1,0, 3), _V(-1,0,0), LNCH_MAXRCSTH, hpr, LNCH_ISP);
	th_rcs[ 9] = CreateThruster (_V(-1,0, 3), _V(1,0,0), LNCH_MAXRCSTH, hpr, LNCH_ISP);
	th_rcs[10] = CreateThruster (_V( 1,0,-3), _V(-1,0,0), LNCH_MAXRCSTH, hpr, LNCH_ISP);
	th_rcs[11] = CreateThruster (_V(-1,0,-3), _V(1,0,0), LNCH_MAXRCSTH, hpr, LNCH_ISP);
	th_rcs[12] = CreateThruster (_V( 0,0,-3), _V(0,0,1), LNCH_MAXRCSTH, hpr, LNCH_ISP);
	th_rcs[13] = CreateThruster (_V( 0,0, 3), _V(0,0,-1), LNCH_MAXRCSTH, hpr, LNCH_ISP);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[2];
	th_group[2] = th_rcs[5];
	th_group[3] = th_rcs[7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_PITCHUP);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[3];
	th_group[2] = th_rcs[4];
	th_group[3] = th_rcs[6];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_PITCHDOWN);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[4];
	th_group[2] = th_rcs[3];
	th_group[3] = th_rcs[7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_BANKLEFT);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[5];
	th_group[2] = th_rcs[2];
	th_group[3] = th_rcs[6];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_BANKRIGHT);

	th_group[0] = th_rcs[0];
	th_group[1] = th_rcs[4];
	th_group[2] = th_rcs[2];
	th_group[3] = th_rcs[6];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_UP);

	th_group[0] = th_rcs[1];
	th_group[1] = th_rcs[5];
	th_group[2] = th_rcs[3];
	th_group[3] = th_rcs[7];
	CreateThrusterGroup (th_group, 4, THGROUP_ATT_DOWN);

	th_group[0] = th_rcs[8];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWLEFT);

	th_group[0] = th_rcs[9];
	th_group[1] = th_rcs[10];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_YAWRIGHT);

	th_group[0] = th_rcs[8];
	th_group[1] = th_rcs[10];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_LEFT);

	th_group[0] = th_rcs[9];
	th_group[1] = th_rcs[11];
	CreateThrusterGroup (th_group, 2, THGROUP_ATT_RIGHT);

	CreateThrusterGroup (th_rcs+12, 1, THGROUP_ATT_FORWARD);
	CreateThrusterGroup (th_rcs+13, 1, THGROUP_ATT_BACK);

	// visual specs
	//AddMesh ("LaunchMK1");
}

// ==============================================================
// API callback interface
// ==============================================================

// --------------------------------------------------------------
// Vessel initialisation
// --------------------------------------------------------------
DLLCLBK VESSEL *ovcInit (OBJHANDLE hvessel, int flightmodel)
{
	return new LNCH (hvessel, flightmodel);
}

// --------------------------------------------------------------
// Vessel cleanup
// --------------------------------------------------------------
DLLCLBK void ovcExit (VESSEL *vessel)
{
	if (vessel) delete (LNCH*)vessel;
}

void LNCH::clbkPostCreation (void)
{
	Fish=ConnectToOrbiterSoundDLL3(GetHandle());

	
	RequestLoadVesselWave3(Fish,1,"Sound\\Rho1\\Orca\\Orca1.wav",RADIO_SOUND);
	SoundOptionOnOff3(Fish,PLAYCOUNTDOWNWHENTAKEOFF,FALSE);
	SoundOptionOnOff3(Fish,DISPLAYTIMER,FALSE);
	SoundOptionOnOff3(Fish,PLAYRADIOATC,FALSE);
	SoundOptionOnOff3(Fish,PLAYATTITUDETHRUST,FALSE);
	SoundOptionOnOff3(Fish,PLAYGPWS,FALSE);
	
}


void LNCH::clbkPreStep (double simt, double simdt, double mjd) 
{
 ///////////////////////////////////////////////////////////////////
       // position relative to object
      VESSEL *v = oapiGetFocusInterface();// to determine which vehicle has focus  
      MinDist1 =  _V(500,500,500); //Distance to trigger sound
      hvessel=oapiGetVesselByName("ZeRG-Explorer");// name of vessel to detect distance (shows in F3 list)
	     if (oapiIsVessel(hvessel) == true) { // check if vessel exist
		  v->GetRelativePos (hvessel, RelPos1);//crashes if hvessel isnt in the senario
	  		                                     //
// change numbers to positive
       if (RelPos1.x < 0.0){
 	RelPos1.x = -(RelPos1.x) ;}
       if (RelPos1.y < 0.0){
    RelPos1.y = -(RelPos1.y) ;}
       if (RelPos1.z < 0.0){ 
    RelPos1.z = -(RelPos1.z) ;} 
   






     if (MinDist1.x > RelPos1.x && MinDist1.y > RelPos1.y && MinDist1.z > RelPos1.z){

		 if (Play_Launch1 < 1){ // play once !
    
			 PlayVesselRadioExclusiveWave3(Fish,1,255); 
			 Play_Launch1 = 1;


		 }// play once
	 }

	
		 } }