// ======================================================================
//                     ORBITER SOFTWARE DEVELOPMENT KIT
//                  Copyright (C) 2001-2004 Martin Schweiger
//                           All rights reserved
// CelBodyAPI.h
// - CELBODY class interface
// - This is the base class for celestial body module implementations
//   (central stars, planets, moons, asteroids, etc.) Plugin modules
//   for celestial bodies define their functionality by deriving a class
//   from CELBODY and overloading the appropriate callback member
//   functions.
// - This class interface replaces the previous interface consisting of
//   global opcXXX callback functions.
// ======================================================================

#ifndef __CELBODYAPI_H
#define __CELBODYAPI_H

// Ephemeris data format bitflags
#define EPHEM_TRUEPOS     0x01
#define EPHEM_TRUEVEL     0x02
#define EPHEM_BARYPOS     0x04
#define EPHEM_BARYVEL     0x08
#define EPHEM_BARYISTRUE  0x10
#define EPHEM_PARENTBARY  0x20
#define EPHEM_POLAR       0x40

// Used for ephemeris interpolation
struct Sample {
	double t;
	double rad;
	double param[6];
};

// ======================================================================
// class CELBODY
// This is the base class for celestial body classes.
// ======================================================================

class DLLEXPORT CELBODY {
public:
	CELBODY ();

	inline int Version() const { return version; }

	// class capabilities
	virtual bool bEphemeris() const;
	// return true if clbkEphemeris and clbkFastEphemeris are implemented

	virtual void clbkInit (FILEHANDLE cfg);
	// initialise instance; provides handle to config file to read parameters

	virtual int clbkEphemeris (double mjd, int req, double *ret);
	// return body's ephemerides (ecliptic, relative to parent) in ret (either
	// in polar or in cartesian coordinates).
	// ret[0-2]: true body position       \
	// ret[3-5]: true body velocity        | if requested in req and
	// ret[6-8]: barycentric position      | supported by module
	// ret[9-11]: barycentric velocity    /
	// return value indicates which requests could be satisfied

	virtual int clbkFastEphemeris (double mjd, int req, double *ret);
	// fast version of the ephemerides calculation used for continuous update
	// (will usually involve interpolation between samples)

	virtual bool clbkAtmParam (double alt, ATMPARAM *prm);
	// return atmospheric parameters (temperature, pressure, density) at
	// altitude alt, if applicable

protected:
	void Pol2Crt (double *pol, double *crt);
	// Convert from polar to cartesian coordinates

	short version;
};

#endif // !__CELBODYAPI_H